#include "Splines.h"
#include <GL/glut.h>
#include <iostream.h>
#include <algorithm>	//needed for the STL copy command

/*
 *	Draw the spline using control points i through i-3,
 *	using the notation from Foley, et al., pg. 505
 *	@param currentLevel The level that we are drawing
 */
void Splines::drawSpline(int currentLevel)
{
	drawControlPoints(currentLevel);
	drawCatmullRom();		// draw the spline between the control points
}

/*
 *	Add another control point to our level
 *	@param currentLevel The level that we want to add our control point.
 */
void Splines::addControlPoint(int x, int y)
{
	// convert the window coordinates
	float xcoord = (x / (float)WindowWidth) * 10. - 5.;
	float ycoord = 5. - (y / (float)WindowHeight) * 10.;
	Vec3 temp(xcoord, ycoord, level);
	
	// store the position of the node
	levelCurves[currentCurve].controlPoints.push_back(temp);
}
/*	Delete a control point from the level
 */
void Splines::deleteControlPoint()
{

	int objectNumber = objectPicked;
	int curve;
	for(int i = 0; i < levelCurves.size(); i++)
	{
		if (levelCurves[i].controlPoints.size() > objectNumber)
		{
			curve = i;
			break;
		}
		else objectNumber -= levelCurves[i].controlPoints.size();
	}
	// shift everything after the object that we picked to the left by one
	for(i = objectNumber; i < levelCurves[curve].controlPoints.size()-1; i++)
	{
		levelCurves[curve].controlPoints[i] = levelCurves[curve].controlPoints[i+1];
	}
	// resize the controlPoints to reflect the loss of the node
	levelCurves[curve].controlPoints.resize(levelCurves[curve].controlPoints.size()-1);
}
/*
 *	Draws the cubes that represent the control points
 */
void Splines::drawControlPoints(int currentLevel)
{
	int nameNumber = 1;
	// cycle through each curve on the level
	for(int j = 0; j<levelCurves.size(); j++)
	{
		// cycle through each node in the curve j
		for (int i = 0; i<levelCurves[j].controlPoints.size(); i++)
		{
			glPushMatrix();
			glTranslatef(levelCurves[j].controlPoints[i].pos[0], 
						 levelCurves[j].controlPoints[i].pos[1], 
						 levelCurves[j].controlPoints[i].pos[2]);
			
			// we are in picking mode, so store the names of the cubes
			if(renderMode == GL_SELECT && currentLevel == level)		
			{
				glLoadName(nameNumber);
				nameNumber++;
			}
			// This is the level and the curve we are focused on
			if(currentLevel == level && currentCurve == j)
			{
				glColor3f(1, 1, 0);
				if(currentLevel==level && currentCurve==j && i==1)glutWireCube(.2);
				glutSolidCube(.1);
				glColor3f(1,0, 0);
			}
			else glutSolidCube(.1);
			glPopMatrix();
		}
	}
}

/*
 *	Draw the splines on the current level.
 *	@param currentLevel The current level we are on.
 */
void Splines::drawCatmullRom()
{
	float newPoint[3];
	int size;

	for(int j = 0; j<levelCurves.size(); j++)
	{
		if(levelCurves[j].controlPoints.size() >= 4 && !levelCurves[j].closure)
		{
			glBegin(GL_LINE_STRIP);
			// The second control point is the first one that has part of the curve through it
			glVertex3fv(levelCurves[j].controlPoints[1].pos);		

			// go through all of the control points
			for(int i=3; i<levelCurves[j].controlPoints.size(); i++)	
			{
				// go at increments of .1 along the curve segments
				for(float dist=.1; dist<1; dist+=.1)				
				{
					interpolatedPoint(dist, i, newPoint, j);
					glVertex3fv(newPoint);		// draw the splined points
				}
			// the next to last control point is the last one to get the curve with catmull-rom
				glVertex3fv(levelCurves[j].controlPoints[i-1].pos);	
			}
			glEnd();		// end of line strip
		}
		else
		{
			size = levelCurves[j].controlPoints.size();
			glBegin(GL_LINE_STRIP);
			glVertex3fv(levelCurves[j].controlPoints[size-2].pos);
			for (int i = 0; i < size; i++)
			{
				for(float dist = .1; dist < 1; dist+=.1)
				{
					interpolatedPoint(dist, i, newPoint, j);
					glVertex3fv(newPoint);
				}
				
				glVertex3fv(levelCurves[j].controlPoints[(i+size-1) % size].pos);
			}
			glEnd();
		}
	}

}
/*
 *	Get the splined points in between the control points
 *	@param t The distance along this specific curve segment
 *	@param i The number of the control point that is the last one that affects this segment of the curve
 *	@param newPoint	Our newly splined point will be kept in here.
 *	@param currentLevel	The level that we are currently on.
 */
void Splines::interpolatedPoint(float t, int i, float newPoint[], int curve)
{
	int size = levelCurves[curve].controlPoints.size();

	// The curve segment goes through points P2 and P3
	int fourth = i;
	int third = (fourth + size-1) % size;
	int second= (third  + size-1) % size;
	int first = (second + size-1) % size;
	float P4[2]={levelCurves[curve].controlPoints[fourth].pos[0], 
		     levelCurves[curve].controlPoints[fourth].pos[1]};
	
	float P3[2]={levelCurves[curve].controlPoints[third].pos[0], 
		     levelCurves[curve].controlPoints[third].pos[1]};
	
	float P2[2]={levelCurves[curve].controlPoints[second].pos[0], 
		     levelCurves[curve].controlPoints[second].pos[1]};
	
	float P1[2]={levelCurves[curve].controlPoints[first].pos[0], 
		     levelCurves[curve].controlPoints[first].pos[1]};

	float ts = t*t;			// t squared
	float tc = ts*t;		// t cubed

	newPoint[0] = .5*((-P1[0]+3*P2[0]-3*P3[0]+P4[0])*tc + 
			(2*P1[0]-5*P2[0]+4*P3[0]-P4[0])* ts + (-P1[0] + P3[0]) * t + 2*P2[0]);
	newPoint[1] = .5*((-P1[1]+3*P2[1]-3*P3[1]+P4[1])*tc + 
			(2*P1[1]-5*P2[1]+4*P3[1]-P4[1])* ts + (-P1[1] + P3[1]) * t + 2*P2[1]);
	newPoint[2] = level;
}

/*
 *	Copy constructor for the Splines class.
 *	@param init	The Spline object that we are copying
 */
Splines::Splines(const Splines &init)
{
	renderMode = init.renderMode;
	level = init.level + 1;
	WindowWidth = init.WindowWidth; WindowHeight = init.WindowHeight;
	currentCurve=init.currentCurve;
	objectPicked = -1;
	
	Nodes node;

	for (int i = 0; i < init.levelCurves.size(); i++)
	{
		levelCurves.push_back(node);
		if(init.levelCurves[i].controlPoints.size()!=0)
		{
			levelCurves[i].controlPoints.insert(levelCurves[i].controlPoints.begin(), 
							init.levelCurves[i].controlPoints.begin(), 
							init.levelCurves[i].controlPoints.end());
			levelCurves[i].closure = init.levelCurves[i].closure;

			for(int j=0; j<levelCurves[i].controlPoints.size(); j++)
				levelCurves[i].controlPoints[j].pos[2]=init.level+1;
		}
	}

}
